% Sample code for J Meng, H Riecke, 2021. 
% The code is also used for other purpose, so it may include unrelavant parameters.
% Ediited by J Meng, 02/23/2021. Contact: John.Meng@nyu.edu

function bulbpiri

Vars=whos;
PersistentVars=Vars([Vars.persistent]);
PersistentVarNames={PersistentVars.name};
clear(PersistentVarNames{:});

run_num_for_saving_figs = dlmread('fignum.txt');
dlmwrite('fignum.txt',run_num_for_saving_figs+1);

% determine the computer on which this is running
[~,hostname]=system('hostname');
fprintf(' Running on %s \n',hostname);
% Mthresh_syn=-0.4;
% alpha_M=0.1;
Mthresh_syn=0;%0.5; % added possibly non-zero threshold in M for resilience of synapse 
alpha_M=5e-3; % prefactor of M in synaptic resilience  % Maybe from now on alpha= 1e3* number of odor.
% ts_syn= 4;%8;%5;%[1.0150/4];         % divide by 4 since that is the number of odors we are using for learning, relevant for repetition_factor. there is an optimum between 0 and 80
G1= 2;%16	 % Explore between 0 and 11
% gamma_syn= 1;%50     % test values between 0.1 and 50
% lambda_R = 1;        % resilience update rate (between 0 and 1)
connM= 30;%[16];       % number of MC's to connect per GC
scaleCS= 0.6;  % scaling factor for coupling strength
syn_age_thr= 0; % threshold in age for spine to actually provide inhibition
% Pr_survive = 0.1;  
S_spontaneous= 0.1;

parameters= setup_parameters(...
    G1,...
    connM,...
    scaleCS, syn_age_thr,...    
    Mthresh_syn,alpha_M,S_spontaneous); % HR

n_parameters= size(parameters,2);

if strcmpi(hostname(1:5),'qnode')     % Used for running on a cluster.
    if n_parameters>1
        parpool(min(n_parameters,24))
        parfor ipar=1:n_parameters
            neurogenesis(parameters(:,ipar),...
                strcat('R',num2str(run_num_for_saving_figs),'_p',num2str(ipar,'%02i')));
        end
    else
            neurogenesis(parameters(:,1),...
                strcat('R',num2str(run_num_for_saving_figs),'_p',num2str(1,'%02i')));
    end
else
    for ipar=1:n_parameters
        neurogenesis(parameters(:,ipar),...
            strcat('R',num2str(run_num_for_saving_figs),'_l',num2str(ipar,'%02i')));
    end
end

    
    function parameters = setup_parameters(varargin)
        ncombi=1;
        for i=1:nargin
            ncombi=ncombi*length(varargin{i});
        end
        parameters=nan(nargin,ncombi);
        outerrep=1;
        innerrep=ncombi;
        for i=1:nargin
            vec=varargin{i};
            nelem=length(vec);
            innerrep=innerrep/nelem;
            temp=nan(1,nelem*innerrep);
            for j=1:nelem
                temp( ((j-1)*innerrep+1):j*innerrep)=vec(j)*ones(1,innerrep);
            end
            parameters(i,:)=repmat(temp(1:nelem*innerrep),1,outerrep);
            outerrep=outerrep*nelem;
        end
    end
end

function neurogenesis(parameters,file_tag)
     N_mitral= 240;
    number_GC_aim_final= 1000;%2000;
% Output folder setup
    mkdir('Saved_Research',file_tag);
    file_location = ['Saved_Research/',file_tag,'/'];
    file_name = strcat(file_location,'bulbpiri_',file_tag,'.m');
    copyfile('bulbpiri.m',file_name);
    para_name= strcat(file_location,'parameter_',file_tag,'.txt');
    fid_parameter=fopen(para_name,'a');
    fprintf(fid_parameter,' Gthresh_syn=%g connM=%g \n scaleCS=%g syn_age_thr=%g Mthresh_syn=%g alpha_M=%g\n  S_spont=%g \n\n',...
    parameters(1),parameters(2),parameters(3),parameters(4),parameters(5),...
    parameters(6),parameters(7) );
    
    % Initializing..
    % Unzip Input parameters
    G1=parameters(1);
%     Mthresh_syn=parameters(5);   %change to the calculated Mthresh.
    alpha_M=parameters(6);
    scaleCS=parameters(3);
    syn_age_thr=parameters(4);
    S_spontaneous=parameters(7);
    file_name = strcat(file_location,'para_',file_tag,'.mat'); 
    % Built-in parameters
    ScaleMitral=120/N_mitral;
    connM=parameters(2)/ScaleMitral;
    
    N_granule=number_GC_aim_final;        % Set dif var for potential neurogenesis.
    CS= 0.01;
    

     CS= CS*(800/800)*(4/24)*scaleCS;
     GSpontAct=0.2;    
     Gthresh=S_spontaneous*24-GSpontAct;

    Mthresh_syn=0;
        G0_syn=0.5; 

    save( file_name,'G1','Mthresh_syn','alpha_M','connM','scaleCS','syn_age_thr','S_spontaneous');
    
  
    tmax=3009;   % First 9 timestep is used for debugging.
    fprintf(' Set maximal time tmax = %g\n',tmax)
    dt= 1;
    n_steps=ceil(tmax/dt);
    
    Dead_num_vec=nan(n_steps,1);
    Add_num_vec=nan(n_steps,1);

    [~,hostname]=system('hostname');

    % Initialize functions, set default value for persistent paras.
    CalGc(0,0,Gthresh,ScaleMitral);
    P_rate(0,0,0,0,Gthresh,CS,ScaleMitral);
    cal_synapses_res(0,0,0,alpha_M,Mthresh_syn,G0_syn,CS);

    synaptic_fluctuation(0,0,connM)
    % learning set configuration
    learning_case=1;
  
    switch learning_case
   %% All kinds of cases
       case 1 
            learning_time=cell(1,4); widths_learn=cell(1,4);
            amplitudes_learn=cell(1,4);odor_nums_learn=cell(1,4);
            xpositions_learn=cell(1,4);
            stim_type_learn=cell(1,4);
            stim_type_learn{1}=1;
            stim_type_learn{2}=1;
            stim_type_learn{3}=1;
            stim_type_learn{4}=1;

            iset=1; learning_time{iset}=0;  % For debugging

            xpositions_learn{iset}=[25;60];
            widths_learn{iset}=[24;24];
            amplitudes_learn{iset}=(2)*[0;0];
            odor_nums_learn{iset}=2;     
            
            iset=2; learning_time{iset}=10;
            xpositions_learn{iset}=[35,85;35,85];   %odor 3,4  
            widths_learn{iset}=[12,12;12,12];
            amplitudes_learn{iset}=(2)*[1,0;0,1];

            odor_nums_learn{iset}=2;     

  
            iset=3; learning_time{iset}=1510;

            xpositions_learn{iset}=[45,75;45,75];   %odor 3,4  

            widths_learn{iset}=[12,12;12,12];
%             amplitudes_learn{iset}=(2)*[1,0;0,1];   % Easy
            amplitudes_learn{iset}=(2)*[0.6,0.4;0.4,0.6];  % Hard
            
            odor_nums_learn{iset}=2;        
            
            iset=4; learning_time{iset}=3010;
            xpositions_learn{iset}=[45,75;45,75];   %odor 3,4  
            widths_learn{iset}=[12,12;12,12];
            amplitudes_learn{iset}=(2-S_spontaneous)*[0.6,0.4;0.4,0.6];
            odor_nums_learn{iset}=2;    

       case 3 % Real Stimuli. Data from Leon's lab cannot be shared. Deleted in the sample code. Available upon request.

    end
    n_learning_set=size(learning_time,2);learning_time{n_learning_set+1}=Inf;

    probe_case=learning_case;
    switch probe_case
        case 1
                iset=1;
                probe_times{iset}=sort([50:50:tmax,[1:5:50],learning_time{2}+[-1,0:5:100],learning_time{3}+[-1,0:5:100],learning_time{4}+[-1,0:5:100],tmax]);
                n_probe_times{iset}=length(probe_times{iset});
                stim_type_probe{iset}=1;
                

                xpositions_probe{iset}=[45,75;45,75;35,85;35,85;50,70];
                widths_probe{iset}=    [12,12;12,12;12,12;12,12;12,12];


%                 amplitudes_probe{1}=(2)*[1,0;0,1;1,0;0,1;0,0];    % Easy
                amplitudes_probe{1}=(2)*[0.6,0.4;0.4,0.6;1,0;0,1;0,0]; % Hard
                odor_nums_probe{iset}=5;

       case 3

                
                
                
    end
    
    state=cell(n_probe_times{iset},8);   % Here the 1D should be the sum of different sets.
    n_probe_set= size(stim_type_probe{1},2);
    fprintf(fid_parameter,' \n learning_case = %g probe_case= %g \n',learning_case,probe_case);
    fclose(fid_parameter);
    
    % Set up input, learn and probe.
    Sall=cell(1,n_learning_set);S_set=cell(1,n_learning_set);
        % Each cell contains [N_mitral,odor_nums_learn] elements
    for learning_set=1:n_learning_set     % Now add a cap.
        if stim_type_learn{learning_set}==3
             S_set{learning_set}= gara3(xpositions_learn{learning_set},...
                widths_learn{learning_set},...
                amplitudes_learn{learning_set},...
                odor_nums_learn{learning_set},...
                stim_type_learn{learning_set},N_mitral);
        else
             Sall{learning_set}= gara3(xpositions_learn{learning_set}/ScaleMitral,...
                widths_learn{learning_set}/ScaleMitral,...
                amplitudes_learn{learning_set},...
                odor_nums_learn{learning_set},...
                stim_type_learn{learning_set},N_mitral);
             S_set{learning_set}=Sall{learning_set}+S_spontaneous;

        end
    end
    S_probe=cell(1,n_probe_set); probe_all=cell(1,n_probe_set);
    for i_probe_set=1:n_probe_set   % Now add a cap.
        if stim_type_learn{learning_set}==3
             S_probe{i_probe_set}= gara3(xpositions_probe{i_probe_set},...
                widths_probe{i_probe_set},...
                amplitudes_probe{i_probe_set},...
                odor_nums_probe{i_probe_set},...
                stim_type_probe{i_probe_set},N_mitral);
        else
            [probe_all{i_probe_set}]= gara3(xpositions_probe{i_probe_set}/ScaleMitral,...
                widths_probe{i_probe_set}/ScaleMitral,...
                amplitudes_probe{i_probe_set},...
                odor_nums_probe{i_probe_set},...
                stim_type_probe{i_probe_set},N_mitral);
            S_probe{i_probe_set} = S_spontaneous+(probe_all{i_probe_set});
        end
    end
    
    % Set up network. We can set random seed from here.
    Wmg=zeros(N_mitral,N_granule);
    Syn_age=nan(N_mitral,N_granule);
%     Syn_res=zeros(N_mitral,N_granule);
    for i_granule=1:N_granule        % Using a random background network.
        temp_Mind=randperm(N_mitral,connM);
        Wmg(temp_Mind,i_granule)=1;
        Syn_age(temp_Mind,i_granule)=0;   %spconvert
    end
    % Initializing
    time=0;
    i_learning_set=1;
    S=S_set{i_learning_set};
    Mc_act=zeros(N_mitral,odor_nums_learn{i_learning_set});    
    Gc_act=zeros(N_granule,odor_nums_learn{i_learning_set});   
    Gc_thr=ones(N_granule,1)*G1;  % For BCM model. Changing based on every odor.
    i_probe_time=1;
    
    Dead_cell_age=nan(1,round(tmax/dt)*connM*N_granule);
    Dead_cell_ind=0;
    % Main loop;
    while (time < tmax)
        time=time+dt;
        % Learn a new set when it's time
        if 1==0   % Here add granule cell for certain condition
            N_granule=N_granule+1;
            % Initialize
%             cal_synapses_res(0,0,alpha_M,Mthresh_syn,Gthresh_syn,N_mitral,N_granule);
            Wmg=[Wmg,zeros(N_mitral,1)];
            Syn_age=[Syn_age,nan(N_mitral,1)];
            temp_Mind=randperm(N_mitral,connM);
            Wmg(temp_Mind,N_granule)=1;
            Syn_age(temp_Mind,N_granule)=0; 
            Gc_act=[Gc_act;zeros(1,odor_nums_learn{i_learning_set})];    

        end
        
        
        if (time >= learning_time{i_learning_set+1})
            i_learning_set=i_learning_set+1;
            S=S_set{i_learning_set};
            fprintf(' \n Mthresh_syn %g, alpha_M %g. learning_set = %g time = %g \n',Mthresh_syn,alpha_M,i_learning_set,time);
            Mc_act=zeros(N_mitral,odor_nums_learn{i_learning_set});    % update mitral cell information.
            Gc_act=zeros(N_granule,odor_nums_learn{i_learning_set});      % update mitral cell information.

        end
       
            % for each odor, calculate the related Mitral cell activity,
            % granule cell activity and other things related to that.
            % The output and related function can change accordingly.
        Syn_res=zeros(N_mitral,N_granule);    
        
        i_odor= randi(odor_nums_learn{i_learning_set}); % Randomly choose the training odor 

            % Calculate Mitral cell activity fix point.
            Mc_act_i=cal_activity(S(:,i_odor),Mc_act(:,i_odor),Wmg);
                    % Stimulus, Initial, Connection
            Mc_act(:,i_odor)=Mc_act_i;
%             Gc_act(:,i_odor)=Wmg'*max(Mc_act_i,0)*ScaleMitral;
            Gc_act(:,i_odor)=CalGc(Mc_act_i,Wmg);
            % Resilience part.
            Syn_res_i= cal_synapses_res(Mc_act_i,Wmg,Gc_thr);
            Syn_res=Syn_res+Syn_res_i;
               

        [Wmg_new,Wmg_dead,Wmg_alive,Wmg_add]=synaptic_fluctuation(Syn_res/odor_nums_learn{i_learning_set},Wmg);
    
        Syn_age=Syn_age+ Wmg_alive*dt;
        
        Dead_cell_num=size(Syn_age(Wmg_dead==1),1);
        Dead_cell_age(Dead_cell_ind+1:Dead_cell_ind+Dead_cell_num)=Syn_age(Wmg_dead==1);
        Dead_cell_ind=Dead_cell_ind+Dead_cell_num;
        Syn_age(Wmg_dead==1)=nan;
        Syn_age(Wmg_add==1)=0;
        Dead_num_vec(round(time/dt))=sum(sum(Wmg_dead));
        Add_num_vec(round(time/dt))=sum(sum(Wmg_add));
        Wmg=Wmg_new;
        
        % Probing part. This part is unnecessary for the main loop. But
        % it's good for digestion.
        if time >= probe_times{1}(i_probe_time)
            fprintf(' \n now probing_set = %g time = %g \n',i_probe_time,time);

            % saving probing related states.
            Mc_prob=zeros(N_mitral,odor_nums_probe{1});    % update mitral cell information.
            Gc_prob=zeros(N_granule,odor_nums_probe{1});  
            
            for i_odor=1:odor_nums_probe{1}
            % Calculate Mitral cell activity fix point by probing
                Mc_prob_i=cal_activity(S_probe{1}(:,i_odor),Mc_prob(:,i_odor),Wmg);
                        % Stimulus, Initial, Connection
                Mc_prob(:,i_odor)=max(Mc_prob_i,0);
                Gc_prob(:,i_odor)=CalGc(Mc_prob_i,Wmg);
            end
                        
            % saving learning related states
            state{i_probe_time,1}= time;
            state{i_probe_time,2}= Mc_act;
            state{i_probe_time,3}= Gc_act;
            state{i_probe_time,4}= Wmg;
            state{i_probe_time,5}= Mc_prob;
            state{i_probe_time,6}= Gc_prob;
            state{i_probe_time,7}= Syn_age;
            state{i_probe_time,8}= Syn_res;
            state{i_probe_time,9}= S;
            state{i_probe_time,10}= S_probe{1};
            state{i_probe_time,11}=Syn_age;
            

            % Save here for intermediate result.
%             figname= 'states';
%             file_name= strcat(file_location,figname,'_',file_tag,'.mat');
%             save(file_name,'state');
            i_probe_time=i_probe_time+1;


        end

    end
    % Save here for speed.
    state=state(1:i_probe_time-1,:);  %#ok<NASGU> I save this variable.
    figname= 'states';
    file_name= strcat(file_location,figname,'_',file_tag,'.mat');
    save(file_name,'state');
    
    % save if synapses distribution required.
    if 1==1 

         figname= 'DeadCell';
         file_name= strcat(file_location,figname,'_',file_tag,'.mat');
         save(file_name,'Dead_cell_ind','Dead_cell_age','Dead_num_vec','Add_num_vec')
    end
    
if ~strcmpi(hostname(1:5),'qnode')  % make a sound when the code is done
    beep;
end
end

function S = gara3(xpositions_learn,widths_learn,amplitudes_learn,odor_nums_learn,stim_type,N_mitral)

odor_redu=nan(N_mitral,odor_nums_learn);

    if stim_type==1
       for is=1:odor_nums_learn
          xpositions=xpositions_learn(is,:); widths=widths_learn(is,:); amplitudes=amplitudes_learn(is,:);
%           odor_redu(:,is)=Flat_mixture(amplitudes,xpositions,widths,N_mitral);
          odor_redu(:,is)=gauss_mixture(amplitudes,xpositions,widths,N_mitral);
       end
          
    elseif stim_type==3 
        Spont=0.1;     % Add air input here. Should make no difference.
        OdorRec=[];

        load('OdorRec.mat');           % 120 mitral cells. Data from Leon's data
        odor_redu=zeros(N_mitral,odor_nums_learn);
        for is=1:odor_nums_learn
           widths=widths_learn(is,:); amplitudes=amplitudes_learn(is,:);
           MixNum=length(widths);
           for im=1:MixNum
             odor_redu(:,is)=odor_redu(:,is)+OdorRec(:,widths(im))*amplitudes(im);
           end
                 odor_redu(:,is)=max((1-Spont)*odor_redu(:,is)+Spont,0);
        end
    end    
      
S=odor_redu;

    function stimulus=gauss_mixture(amplitudes,xpositions,widths,N_mitral)
        x=1:N_mitral;
        n_components=length(xpositions);
        stimulus=zeros(1,N_mitral);
        for icomp=1:n_components
            stimulus=stimulus+amplitudes(icomp)*exp(-(x-xpositions(icomp)).^2/widths(icomp)^2);
        end

    end
end

function  Mc=cal_activity(S,Mc_ini,Wmg)
    options = odeset('RelTol',1e-4,'AbsTol',1e-5','OutputFcn',@out_fun2);
    tmax = 500;
    
    lastP=Mc_ini;
    [~,Y]=ode45(@(t,y) P_rate(t,y,Wmg,S),[0,tmax],Mc_ini,options);
    Mc=Y(end,:)';
   
    function stop_ = out_fun2(~,P_,flag_)
        stop_ = false;
        if strcmp(flag_,'init')
        elseif strcmp(flag_,'done')
        else
            if size(P_,2) > 1
                cond_ = norm(lastP-P_(:,end));
            else
                cond_ = norm(lastP-P_);
            end
            %if cond_<1e-2
            if cond_<1e-4
                stop_ = true;
            end
            if size(P_,2) > 1
                lastP = P_(:,end);
            else
                lastP = P_;
            end
        end
    end % out_fun
end

function dY=P_rate(~,Y,Wmg,stim_select,varargin) %differential equation
persistent  Gthresh CS ScaleMitral  % Maybe Gthresh need to be reset. ???
if nargin>4
    Gthresh=varargin{1};
    CS=varargin{2};
    ScaleMitral=varargin{3};
    return;
end

Mc=Y; 
sigMc=max(Mc,0);

GcTemp = ((Wmg')*sigMc)*ScaleMitral;

Gc= max(GcTemp-Gthresh,0);
dM= -Mc+max(tanh(stim_select-(CS*Wmg)*Gc),0);


dY=dM;

end 

function Res=cal_synapses_res(Mc,Wmg,G1,varargin)
persistent alpha_M Mthresh_syn G0_syn CS  %N_mitral N_granule
                        % Granule cell mark.
if nargin>3
    alpha_M=varargin{1};
    Mthresh_syn=varargin{2};
    G0_syn=varargin{3};   % Here it means the G_normalize factor. I add G1, Gman here for now. BUT it should come from outside.
    CS=varargin{4};

    Res=0;
    return;
end

    Gc= CalGc(Mc,Wmg);     
    Mc_rec=max(Mc-Mthresh_syn,0);    % Maybe we need to change that.  Currently always set to 0.

    Gc_rec=max((Gc-G0_syn),0).*(Gc-G1);
    Res=alpha_M*Mc_rec*Gc_rec';


end

function [Wmg_new,Wmg_dead, Wmg_alive, Wmg_add]=synaptic_fluctuation(Syn_res,Wmg,varargin)
% Wmg_new Wmg_dead, Wmg_alive, Wmg_add are all matrix. Strategy based on
% forming first and removing second
persistent connM 
if nargin>2
    connM=varargin{1};
    return;
    
end
    [Mc_num,Gc_num]=size(Wmg);          
    
    lambda=(Mc_num-connM)/connM;
    eps2=0;
    eps1=eps2*lambda/10;
    
% Dying because of competition
    Gc_ind=1:Gc_num;
    Wmg_new=Wmg>0;
    Syn_num=sum(Wmg,1);
    if ~all(Syn_num<=connM*1.1)   % Set the upper bound.  ???? connM*1.5
        Gc_vec= Gc_ind(Syn_num>connM*1.1);
        for Gc_i=Gc_vec
            temp_res=Syn_res(Wmg_new(:,Gc_i),Gc_i);
            temp_ind=find(Wmg_new(:,Gc_i)>0);
            [~,Mc_kl]= maxk(temp_res,round(connM*1.1));
            temp_vec=zeros(Mc_num,1);
            temp_vec(temp_ind(Mc_kl))=1;
            Wmg_new(:,Gc_i)=temp_vec;
        end
    end
    
    Wmg_temp=Wmg_new;    
    Wmg_dead_temp=(1-Wmg_temp).*Wmg;
    
    Syn_res_cal= (Syn_res+ Wmg*(-eps1)+(1-Wmg)*eps2);   % eps for spontaneous change
    
    Syn_cal=exp(10* Syn_res_cal).*Wmg+(1-exp(-Syn_res_cal)).*(1-Wmg);   % The first term is for forming and the second is for forgeting.
                % 1st term, for the not existing spots, forming rate is..
                % 2nd term, for the existing spines, removing rate is 10x
                % normal

    RandM=rand(Mc_num,Gc_num);
    Flag= Syn_cal + RandM;
    Syn_fate = Flag>=1;   % 1 survive 0 dead
    Wmg_new=Syn_fate.*(1-Wmg_dead_temp);
    
    Syn_num_new=sum(Wmg_new,1);            %b
    Wmg_dead= ((1-Wmg_new).*Wmg)>0;
    
    
    Wmg_alive=Wmg_new.*Wmg;
    Wmg_add= Wmg_new.*(1-Wmg);
   
end

function [VecOut,Ind] = maxk(vec,k)
ll=length(vec);

vecT=vec+rand(ll,1)*0.001;
[~,indv]=sort(vecT);
Ind=indv(end:-1:end-k+1);
VecOut=vec(Ind);

end

function Gc = CalGc(Mc,Wmg,varargin)
persistent  ScaleMitral Gthresh
if nargin>2
    Gthresh=varargin{1};
    ScaleMitral=varargin{2};
    return;
end
sigMc=max(Mc,0);    
GcTemp = ((Wmg')*sigMc)*ScaleMitral;
Gc= max(GcTemp-Gthresh,0);

end

